from setuptools import setup, find_packages

# Setting Up

setup(
      
      name = "pyerualjetwork",
      version = "1.0.9",
      author = "Hasan Can Beydili",
      author_email = "tchasancan@gmail.com",
      description= "Advanced python deep learning library.(More document coming soon..)",
      packages = find_packages(),
      keywords = ["model evaluation", "classifcation", 'divided pruning neural networks']
      
      )